---------------------------------------------------------------------------------
-- eu2bytes.e
-- version 0.1a
-- written by Chris Bensler      mailto: bensler@mail.com
---------------------------------------------------------------------------------
--
---------------------------------------------------------------------------------
include std/core.em
include std/mem/binary.e

-- Compressed format of Euphoria objects
-- EC?B,LENGTH,BYTES...
global constant
  ECIU1B = 1,
  ECIU2B = 2,
  ECIU4B = 4,

  ECIS1B = 5, -- 4+1 -- 1-byte integer
  ECIS2B = 6, -- 4+2 -- 2-byte integer
  ECIS4B = 8, -- 4+4 -- 4-byte integer

  ECFP4B = 12, -- 8+4 -- 4-byte f.p.
  ECFP8B = 16, -- 8+8 -- 8-byte f.p.
  
  ECS1B  = 17, -- 16+1 -- sequence, 1-byte length of bytes
  ECS4B  = 20, -- 16+4 -- sequence, 4-byte length of bytes
  
  ECST1B = 33, -- 32+1 -- string, 1-byte length of bytes
  ECST4B = 36  -- 32+4 -- string, 4-byte length of bytes

constant
  MAXU1B  = 255,
  MAXU2B  = power(2,16)-1,
  MINS1B  = -128,
  MINS2B  = -power(2,15),
  MINS4B  = -power(2,31)


---------------------------------------------------------------------------------
global function eu_byte_length(sequence s)
 integer sz
  sz = s[1]
  if sz > 16 then
    if and_bits(sz,1) then -- sequence or string with length <= 255
      sz = 1 + s[2] -- 1 for size, plus s[2] (sequence length)
    else -- sequence or string with size > 255
      sz = 4 + bytes_to_int(s[2..5]) -- 4 for size
    end if
  elsif sz < 12 then
    sz = remainder(sz,4)
    if not sz then sz = 4 end if
  else
    sz -= 8
  end if
  sz += 1 -- add 1 for the 'type' byte
  return sz
end function

---------------------------------------------------------------------------------
global function eu_to_bytes(object x)
 sequence s
  if integer(x) then    
    if x >= 0 then
      if    x <= MAXU1B then  s = {ECIU1B, x}
      elsif x <= MAXU2B then  s = {ECIU2B, lo_byte(x), hi_byte(x)}
      else                    s = {ECIU4B} & int_to_bytes(x)
      end if
    else
      if    x >= MINS1B then  s = {ECIS1B, x-MINS1B}
      elsif x >= MINS2B then  s = {ECIS2B, lo_byte(x-MINS2B), hi_byte(x-MINS2B)}
      else                    s = {ECIS4B} & int_to_bytes(x-MINS4B)
      end if
    end if
  elsif atom(x) then
    s = atom_to_float32(x)
    if x = float32_to_atom(s) then  s = {ECFP4B} & s
    else                            s = {ECFP8B} & atom_to_float64(x)
    end if
  elsif string(x) then
    if length(x) <=255 then   s = {ECST1B} & length(x)
    else                      s = {ECST4B} & int_to_bytes(length(x))
    end if
    s &= x
  else
    s = {}
    for i = 1 to length(x) do
      s &= eu_to_bytes(x[i])
    end for
    if length(s) <= 255 then  s = {ECS1B,length(s)} & s
    else                      s = {ECS4B} & int_to_bytes(length(s)) & s
    end if
  end if
  return s
end function


---------------------------------------------------------------------------------
-- returns {object, len}
--  If len = length(s), the stream was successfully converted, and there are no
--   remaining bytes.
--  If len > length(s), the stream was only partial, and couldn't be converted.
--   You need to append the missing bytes to your stream, and try again.
--  If len < length(s), the stream was successfully converted, and there are
--   remaining bytes in the stream. This usually means there is another object in the stream.
--   You can try to convert the next object, by slicing off the converted portion
--   of the stream, and converting the remaining bytes.
global function bytes_to_eu(sequence s)
 integer l,sl
 object x, tmp
  l = eu_byte_length(s)
  if l > length(s) then return {0,l} end if
  
  if    s[1] = ECIU1B then  x = s[2]
  elsif s[1] = ECIS1B then  x = s[2]+MINS1B
  elsif s[1] = ECIU2B then  x = make_word(s[2],s[3])
  elsif s[1] = ECIS2B then  x = make_word(s[2],s[3])+MINS2B
  elsif s[1] = ECIU4B then  x = bytes_to_int(s[2..5])
  elsif s[1] = ECIS4B then  x = bytes_to_int(s[2..5])+MINS4B
  elsif s[1] = ECFP4B then  x = float32_to_atom(s[2..5])
  elsif s[1] = ECFP8B then  x = float64_to_atom(s[2..9])
  elsif s[1] = ECST1B then  x = s[3..2+s[2]]
  elsif s[1] = ECST4B then  x = s[6..(5 + bytes_to_int(s[2..5]))]
  else
    if s[1] = ECS1B then sl = 3
    else                 sl = 6
    end if
    x = {}
    while sl <= l do
      tmp = bytes_to_eu(s[sl..length(s)])
      x &={ tmp[1] }
      if tmp[2]-1 > length(s)-sl then return {0,tmp[2]-1} end if
      sl += tmp[2]
    end while
  end if
  
  return {x,l}
end function