---------------------------------------------------------------------------------
-- callback.e
-- version 2
-- written by Chris Bensler      mailto: bensler@mail.com
---------------------------------------------------------------------------------
-- a simple mechanism for improving forward referencing
--  We can use an indexed list of names and routine id's, to call forward referenced functions by name.
--    fref(name, id): declare a new forward reference
--    id = fref_id(name): get the id of a forward reference
--    call_proc(id, args): redeclared, to allow for calling an fref()'d routine by name
--    call_func(id, args): redeclared, to allow for calling an fref()'d routine by name
-- * NOTE*
--    You still cannot call a forward reference before it's declared.
--    Any call to a routine that doesn't exist yet will crash, of course.
---------------------------------------------------------------------------------
include std/core/declares.e

---------------------------------------------------------------------------------
-- declare the index of fref names and routine id's
sequence Index      Index={{},{}}
constant
      NAME=1
     ,ID=2


---------------------------------------------------------------------------------
-- GLOBALS --

------------------------------------------------
global procedure define_callback(sequence name, integer id)
------------------------------------------------
-- syntax:        include fref_ids.e
--                fref(name,id)
-- description:
--    Add a new routine id to the forward reference list
------------------------------------------------
 integer f
  f = find(name, Index[NAME])
  if f then
    Index[ID][f] = id
  else
   Index[NAME] &={ name }
   Index[ID]  &={ id }
  end if
end procedure
------------------------------------------------


------------------------------------------------
global function callback_id(sequence name)
------------------------------------------------
-- syntax:        include fref_ids.e
--                id = fref_id(name)
-- returns: a routine id
-- description:
--    This routine retrieves a routine id from the fref indexed list.
--    This is similar to routine_id(), but will only work for routine id's stored with fref()
--     and fref()/fref_id() will work across external includes, without having to define any global vars for the id's
------------------------------------------------
 integer f
   f = find(name, Index[NAME])
   if not f then return -1 end if
   return Index[ID][f]
end function
------------------------------------------------

------------------------------------------------
global function callback_name(integer id)
------------------------------------------------
-- syntax:        include fref_ids.e
--                name = fref_name(id)
-- returns: sequence
-- description:
--    fetch the name of a routine, specified by id, from the fref index
--    if the id is not found in the index, "" will be returned
------------------------------------------------
 integer f
  f = find(id, Index[ID])
  if not f then return "" end if
  return Index[NAME][f]
end function
------------------------------------------------

-------------------------------------------------
-- Wrappers for calling referenced (routine id) eu routines --
--  These wrappers add the ability to call_proc/func() by name, as well as id
--  These only work, if the routine you are calling used fref()

------------------------------------------------
global procedure proc_call(object id, sequence args)
------------------------------------------------
-- syntax:        include fref_ids.e
--                fref_proc(name_or_id, args)
-- description:
--    
------------------------------------------------
   if sequence(id) then id=callback_id(id) end if
   call_proc(id,args)
end procedure
------------------------------------------------

------------------------------------------------
global function func_call(object id, sequence args)
------------------------------------------------
-- SYNTAX:        include fref_ids.e
--                x = fref_func(name_or_id, args)
-- PARAMETERS:
--  id is either a string referring to a routine that was previously stored in the fref table,
--      or if an atom, it should be a routine id
--  args is a sequence of parameters that the callback routine takes
-- description:
--    
------------------------------------------------
   if sequence(id) then id=callback_id(id) end if
   return call_func(id,args)
end function
------------------------------------------------


-- END OF GLOBALS --
---------------------------------------------------------------------------------


---------------------------------------------------------------------------------
-- BUILTIN WRAPPERS --
-- so we can get a routine id for each of the builtin's

function eu_length(sequence s)
   return length(s)
end function

procedure eu_puts(integer fn, object x)
   puts(fn,x)
end procedure

type eu_integer(object x)
   return integer(x)
end type

type eu_atom(object x)
   return atom(x)
end type

type eu_sequence(object x)
   return sequence(x)
end type

without warning
type eu_object(object x)
   return 1
end type
with warning

procedure eu_position(integer row, integer column)
   position(row, column)
end procedure

function eu_append(sequence s, object x)
   return append(s, x)
end function

function eu_prepend(sequence s, object x)
   return prepend(s, x)
end function

procedure eu_print(integer fn, object x)
   print(fn, x)
end procedure

procedure eu_printf(integer fn, sequence st, object x)
   printf(fn, st, x)
end procedure

procedure eu_clear_screen()
   clear_screen()
end procedure

function eu_floor(object x)
   return floor(x)
end function

function eu_getc(integer fn)
   return getc(fn)
end function

function eu_gets(integer fn)
   return gets(fn)
end function

function eu_get_key()
   return get_key()
end function

function eu_rand(object x)
   return rand(x)
end function

function eu_repeat(object x, atom a)
   return repeat(x,a)
end function

function eu_compare(object x1, object x2)
   return compare(x1,x2)
end function

function eu_find(object x, sequence s)
   return find(x,s)
end function

function eu_match(sequence s1, sequence s2)
   return match(s1,s2)
end function

function eu_time()
   return time()
end function

function eu_command_line()
   return command_line()
end function

function eu_open(sequence st1, sequence st2)
   return open(st1,st2)
end function

procedure eu_close(integer fn)
   close(fn)
end procedure

procedure eu_trace(integer i)
   trace(i)
end procedure

function eu_getenv(sequence s)
   return getenv(s)
end function

function eu_sqrt(object x)
   return sqrt(x)
end function

function eu_sin(object x)
   return sin(x)
end function

function eu_cos(object x)
   return cos(x)
end function

function eu_tan(object x)
   return tan(x)
end function

function eu_log(object x)
   return log(x)
end function

procedure eu_system(sequence st, integer i)
   system(st, i)
end procedure

function eu_date()
   return date()
end function

function eu_remainder(object x1, object x2)
   return remainder(x1,x2)
end function

function eu_power(object x1, object x2)
   return power(x1,x2)
end function

function eu_machine_func(atom a, object x)
   return machine_func(a,x)
end function

procedure eu_machine_proc(atom a, object x)
   machine_proc(a,x)
end procedure

procedure eu_abort(integer i)
   abort(i)
end procedure

function eu_peek(object a)
   return peek(a)
end function

procedure eu_poke(atom a, object x)
   poke(a,x)
end procedure

procedure eu_call(atom a)
   call(a)
end procedure

function eu_sprintf(sequence st, object x)
   return sprintf(st,x)
end function

function eu_arctan(object x)
   return arctan(x)
end function

function eu_and_bits(object x1, object x2)
   return and_bits(x1,x2)
end function

function eu_or_bits(object x1, object x2)
   return or_bits(x1,x2)
end function

function eu_xor_bits(object x1, object x2)
   return xor_bits(x1,x2)
end function

function eu_not_bits(object x)
   return not_bits(x)
end function

procedure eu_pixel(object x, sequence s)
   pixel(x,s)
end procedure

function eu_get_pixel(sequence s)
   return get_pixel(s)
end function

procedure eu_mem_copy(atom a1, atom a2, integer i)
   mem_copy(a1,a2,i)
end procedure

procedure eu_mem_set(atom a, integer i1, integer i2)
   mem_set(a,i1,i2)
end procedure

procedure eu_c_proc(integer i, sequence s)
   c_proc(i,s)
end procedure

function eu_c_func(integer i, sequence s)
   return c_func(i,s)
end function

function eu_routine_id(sequence st)
   return routine_id(st)
end function

procedure eu_call_proc(integer i, sequence s)
   call_proc(i,s)
end procedure

function eu_call_func(integer i, sequence s)
   return call_func(i,s)
end function

procedure eu_poke4(atom a, object x)
   poke4(a,x)
end procedure

function eu_peek4s(object x)
   return peek4s(x)
end function

function eu_peek4u(object x)
   return peek4u(x)
end function

procedure eu_profile(integer i)
   profile(i)
end procedure

function eu_equal(object x1, object x2)
   return equal(x1,x2)
end function

function eu_system_exec(sequence st, integer i)
   return system_exec(st,i)
end function

function eu_platform()
   return platform()
end function

-- END OF BUILTIN WRAPPERS --
---------------------------------------------------------------------------------


---------------------------------------------------------------------------------
-- initialize the index with the builtin id's
for i = 1 to length(BUILTINS) do
  define_callback(BUILTINS[i],routine_id("eu_"&BUILTINS[i]))
end for


---------------------------------------------------------------------------------
-- THINGS TO DO --
