include std/core/declares.e
include std/core/callback.e

---------------------------------------------------------------------------------
-- CRASH ROUTINES --

global procedure crash_message(sequence msg)
-- Specify a final message to display for your user, in the event 
-- that Euphoria has to shut down your program due to an error.
    machine_proc(M_CRASH_MESSAGE, msg)
end procedure

global procedure crash_file(sequence file_path)
-- Specify a file path name in place of "ex.err" where you want
-- any diagnostic information to be written.
    machine_proc(M_CRASH_FILE, file_path)
end procedure


---------------------------------------------------------------------------------
-- BASIC ERROR ROUTINES --

-- force a fatal error and display the given msg
-- use this to force a debug dump to ex.err
global procedure fatal_error(sequence msg)
   crash_message(msg)
   ? 1/0
end procedure

global procedure abort_error(integer code,sequence msg)
  if length(msg) then puts(STDERR,msg) end if
  machine_proc(M_WAIT_KEY, NULL)
  abort(code)
end procedure

-- default callback for the error routine, below
define_callback("error_routine" ,routine_id("fatal_error"))

global procedure error(sequence args)
  proc_call("error_routine",{args})
end procedure


---------------------------------------------------------------------------------
-- ADVANCED, ERROR TRAPPING ROUTINES --
-- By defining error id's and associating handlers, errors can be handled almost automatically,
--  while maintaining, and even improving flexibility.
-- These routines make it possible to throw an error from one routine, back through the call stack,
--  where they can be trapped and handled automatically, using the defined error handler,
--  or they can be caught and handled manually at the program level.
-- In more cases than not, error trapping is faster than manual/traditional error handling

sequence error_index        error_index = {}
integer pending_id          pending_id = NULL
object pending_data

-- defines a new error id
global function define_error(integer handler_id)
  if handler_id = NULL then
    handler_id = callback_id("error_routine") -- default error handler
  end if
  error_index &= handler_id
  return length(error_index)
end function

-- redefines an error handler
global procedure redefine_error(integer error_id, integer handler_id)
  if error_id > 0 and error_id <= length(error_index) then
    error_index[error_id] = handler_id
  end if
end procedure

global procedure trap_error()
  if pending_id then
    call_proc(error_index[pending_id],{pending_data})
    pending_id = NULL
  end if
end procedure

global procedure throw_error(integer error_id, object data)
  trap_error()
  pending_id = error_id
  pending_data = data
end procedure

global function catch_error()
  return pending_id
end function

global procedure drop_error()
  pending_id = NULL
end procedure

global function fetch_error()
  return pending_data
end function
