include std/core.e
include std/mem.e

type color(integer x)
    return x >= 0 and x <= 255
end type

type mode(integer x)
  return (x >= -3 and x <= 19) or (x >= 256 and x <= 263)
end type

type page_number(integer p)
    return p >= 0 and p <= 7
end type

type rgb_seq(sequence s)
    return length(s) = 3 -- {red, green, blue}
end type

type point(sequence x)
    return length(x) = 2
end type

type point_seq(sequence x)
    return length(x) >= 2
end type


---------------------------------------------------------------------------------
-- CONSOLE GRAPHICS CONFIGURATION ROUTINES --

global constant -- video_config structure indices
        VC_COLOR    = 1
       ,VC_MODE     = 2
       ,VC_LINES    = 3
       ,VC_COLUMNS  = 4
       ,VC_XPIXELS  = 5
       ,VC_YPIXELS  = 6
       ,VC_NCOLORS  = 7
       ,VC_PAGES    = 8


global function video_config() -- M_VIDEO_CONFIG    = 13    --
-- return sequence of information on video configuration
-- {color?, mode, text lines, text columns, xpixels, ypixels, #colors, pages}
  return machine_func(M_VIDEO_CONFIG, 0)
end function

global procedure use_vesa(boolean code) -- M_USE_VESA        = 36    --
-- If code is 1 then force Euphoria to use the VESA graphics standard.
-- This may let Euphoria work better in SVGA modes with certain graphics cards.
-- If code is 0 then Euphoria's normal use of the graphics card is restored.
-- Values of code other than 0 or 1 should not be used.
  machine_proc(M_USE_VESA, code)
end procedure

global function graphics_mode(mode m) -- M_GRAPHICS_MODE   = 5     --
-- try to set up a new graphics mode
-- return 0 if successful, non-zero if failed
  return machine_func(M_GRAPHICS_MODE, m)
end function

global function get_display_page() -- M_GET_DISPLAY_PAGE= 28    --
-- return current page# mapped to the monitor   
  return machine_func(M_GET_DISPLAY_PAGE, 0)
end function

global procedure set_display_page(page_number page) -- M_SET_DISPLAY_PAGE= 29    --
-- select a page to be displayed
  machine_proc(M_SET_DISPLAY_PAGE, page)
end procedure

global function get_active_page() -- M_GET_ACTIVE_PAGE = 30    --
-- return current page# that screen output is sent to
  return machine_func(M_GET_ACTIVE_PAGE, 0)
end function

global procedure set_active_page(page_number page) -- M_SET_ACTIVE_PAGE = 31    --
-- select a page for screen output
  machine_proc(M_SET_ACTIVE_PAGE, page)
end procedure


---------------------------------------------------------------------------------
-- GRAPHICS PRIMITIVES ROUTINES --

global procedure draw_line(color c, point_seq xyarray) -- M_LINE            = 2     --
-- draw a line connecting the 2 or more points
-- in xyarray: {{x1, y1}, {x2, y2}, ...}
-- using a certain color 
  machine_proc(M_LINE, {c, 0, xyarray})
end procedure

global procedure ellipse(color c, boolean fill, point p1, point p2) -- M_ELLIPSE         = 18    --
-- draw an ellipse with a certain color that fits in the
-- rectangle defined by diagonal points p1 and p2, i.e. 
-- {x1, y1} and {x2, y2}. The ellipse may be filled or just an outline.   
  machine_proc(M_ELLIPSE, {c, fill, p1, p2})
end procedure

global procedure polygon(color c, boolean fill, point_seq xyarray) -- M_POLYGON         = 11    --
-- draw a polygon using a certain color
-- fill the area if fill is TRUE
-- 3 or more vertices are given in xyarray
  machine_proc(M_POLYGON, {c, fill, xyarray})
end procedure


---------------------------------------------------------------------------------
-- PALETTE ROUTINES --

global function palette(color c, rgb_seq s) -- M_PALETTE         = 3     --
-- choose a new mix of {red, green, blue} to be shown on the screen for
-- color number c. Returns previous mixture as {red, green, blue}.
  return machine_func(M_PALETTE, {c, s})
end function

global procedure all_palette(sequence s) -- M_ALL_PALETTE     = 27    --
-- s is a sequence of the form: {{r,g,b},{r,g,b}, ...{r,g,b}}
-- that specifies new color intensities for the entire set of
-- colors in the current graphics mode.  
  machine_proc(M_ALL_PALETTE, s)
end procedure

global function get_all_palette()
-- Get color intensities for the entire set of colors in the current 
-- graphics mode. Returned sequence is {{r,g,b},{r,g,b},...,{r,g,b}}. 
-- Intensity values are in the range 0 to 63.
 integer mem, numColors
 sequence vc, reg, colors
  vc = video_config()
  numColors = vc[VC_NCOLORS]
  reg = repeat(0, REG_LIST_SIZE)
  mem = alloc_low(numColors*3)
  if mem then
  	reg[REG_AX] = #1017
  	reg[REG_BX] = 0
  	reg[REG_CX] = numColors
  	reg[REG_ES] = floor(mem/16)
  	reg[REG_DX] = and_bits(mem, 15)
  	reg = dos_interrupt(#10, reg)
  	colors = {}
  	for col = mem to mem+(numColors-1)*3 by 3 do
	    colors = append(colors, peek({col,3}))
  	end for
  	dealloc_low(mem)
  	return colors
  else
  	return {} -- unlikely
  end if
end function


---------------------------------------------------------------------------------
-- IMAGE ROUTINES --
type coordinates(sequence p)
    return length(p) = 2 and p[1] >= 0 and p[2] >= 0
end type

global procedure display_image(coordinates xy, sequence pixels)
-- display a 2-d sequence of pixels at location xy
-- N.B. coordinates are {x, y} with {0,0} at top left of screen
-- and x values increasing towards the right, 
-- and y values increasing towards the bottom of the screen
  for i = 1 to length(pixels) do
	  pixel(pixels[i], xy)
	  xy[2] = xy[2] + 1
  end for
end procedure

global function save_image(coordinates top_left, coordinates bottom_right)
-- Save a rectangular region on a graphics screen,
-- given the {x, y} coordinates of the top-left and bottom-right 
-- corner pixels. The result is a 2-d sequence of pixels suitable 
-- for use in display_image() above.
 integer x, width
 sequence save
  x = top_left[1]
  width = bottom_right[1] - x + 1
  save = {}
  for y = top_left[2] to bottom_right[2] do
  	save = append(save, get_pixel({x, y, width}))
  end for
  return save
end function


---------------------------------------------------------------------------------
-- CONSOLE TEXT GRAPHICS ROUTINES --

type text_point(sequence p)
    return length(p) = 2 and p[1] >= 1 and p[2] >= 1 
	   and p[1] <= 200 and p[2] <= 500 -- rough sanity check
end type

constant MONO_TEXT_MEMORY   = #B0000
constant COLOR_TEXT_MEMORY  = #B8000

constant BYTES_PER_CHAR = 2

function DOS_scr_addr(sequence vc, text_point xy)
-- calculate address in DOS screen memory for a given line, column
 atom screen_memory
 integer page_size
    
  if vc[VC_MODE] = 7 then
	  screen_memory = MONO_TEXT_MEMORY
  else
	  screen_memory = COLOR_TEXT_MEMORY
  end if
  page_size = vc[VC_LINES] * vc[VC_COLUMNS] * BYTES_PER_CHAR
  page_size = 1024 * floor((page_size + 1023) / 1024)
  screen_memory = screen_memory + get_active_page() * page_size
  return screen_memory + ((xy[1]-1) * vc[VC_COLUMNS] + (xy[2]-1)) * BYTES_PER_CHAR
end function
-------------------------------------------------

global function get_screen_char(positive_atom line, positive_atom column)
-- returns {character, attributes} of the single character
-- at the given (line, column) position on the screen
 atom scr_addr
 sequence vc

  if platform() = DOS32 then
  	vc = video_config()
	  if line >= 1 and line <= vc[VC_LINES] and
	        column >= 1 and column <= vc[VC_COLUMNS] then
	    scr_addr = DOS_scr_addr(vc, {line, column})
	    return peek({scr_addr, 2})
	  else
	    return {0,0}
	  end if
  else    
	  return machine_func(M_GET_SCREEN_CHAR, {line, column})
  end if
end function

global procedure put_screen_char(positive_atom line, positive_atom column,sequence char_attr)
-- stores {character, attributes, character, attributes, ...} 
-- of 1 or more characters at position (line, column) on the screen
 atom scr_addr
 sequence vc
 integer overflow

  if platform() = DOS32 then
  	vc = video_config()
	  if line <= vc[VC_LINES] and column <= vc[VC_COLUMNS] then
	    scr_addr = DOS_scr_addr(vc, {line, column})
	    overflow = length(char_attr) - 2 * (vc[VC_COLUMNS] - column + 1)
	    if overflow > 0 then
		    poke(scr_addr, char_attr[1..length(char_attr) - overflow])  
	    else
		    poke(scr_addr, char_attr)
	    end if
	  end if
  else    
	  machine_proc(M_PUT_SCREEN_CHAR, {line, column, char_attr})
  end if
end procedure

-------------------------------------------------

global procedure display_text_image(text_point xy, sequence text)
-- Display a text image at line xy[1], column xy[2] in any text mode.
-- N.B. coordinates are {line, column} with {1,1} at the top left of screen
-- Displays to the active text page.
 atom scr_addr
 integer screen_width, extra_col2, extra_lines
 sequence vc, one_row

  vc = video_config()
  if platform() = DOS32 then
  	screen_width = vc[VC_COLUMNS] * BYTES_PER_CHAR
	  scr_addr = DOS_scr_addr(vc, xy)
  end if
  if xy[1] < 1 or xy[2] < 1 then
	  return -- bad starting point
  end if
  extra_lines = vc[VC_LINES] - xy[1] + 1 
  if length(text) > extra_lines then
	  if extra_lines <= 0 then
	    return -- nothing to display
	  end if
	  text = text[1..extra_lines] -- truncate
  end if
  extra_col2 = 2 * (vc[VC_COLUMNS] - xy[2] + 1)
  for row = 1 to length(text) do
	  one_row = text[row]
	  if length(one_row) > extra_col2 then
	    if extra_col2 <= 0 then
		    return -- nothing to display
	    end if
	    one_row = one_row[1..extra_col2] -- truncate
	  end if
	  if platform() = DOS32 then
	    poke(scr_addr, one_row)
	    scr_addr += screen_width
	  else
	    -- Linux
	    machine_proc(M_PUT_SCREEN_CHAR, {xy[1]+row-1, xy[2], one_row})
	  end if
  end for
end procedure

global function save_text_image(text_point top_left, text_point bottom_right)
-- Copy a rectangular block of text out of screen memory,
-- given the coordinates of the top-left and bottom-right corners.
-- Reads from the active text page.
 sequence image, row_chars, vc
 atom scr_addr, screen_memory
 integer screen_width, image_width
 integer page_size

  vc = video_config()
  screen_width = vc[VC_COLUMNS] * BYTES_PER_CHAR
  if platform() = DOS32 then
  	if vc[VC_MODE] = 7 then
	    screen_memory = MONO_TEXT_MEMORY
	  else
	    screen_memory = COLOR_TEXT_MEMORY
	  end if
	  page_size = vc[VC_LINES] * screen_width
	  page_size = 1024 * floor((page_size + 1023) / 1024)
	  screen_memory = screen_memory + get_active_page() * page_size
	  scr_addr = screen_memory + 
    		(top_left[1]-1) * screen_width + 
	    	(top_left[2]-1) * BYTES_PER_CHAR
  end if
  image = {}
  image_width = (bottom_right[2] - top_left[2] + 1) * BYTES_PER_CHAR
  for row = top_left[1] to bottom_right[1] do
	  if platform() = DOS32 then
	    row_chars = peek({scr_addr, image_width})
	    scr_addr += screen_width
	  else
	    -- Linux
	    row_chars = {}
	    for col = top_left[2] to bottom_right[2] do
		    row_chars &= machine_func(M_GET_SCREEN_CHAR, {row, col})
	    end for
	  end if
	  image = append(image, row_chars)
  end for
  return image
end function


---------------------------------------------------------------------------------
-- CONSOLE GRAPHICS MOUSE ROUTINES --

sequence tmp
if platform() <= WIN32 then
  tmp = {0,4,4,1,1,2,2
          ,35}  -- ANY_UP: LEFT, RIGHT or MIDDLE up (best you can do under xterm)
else
  tmp = {1,2,4,8,16,32,64
          ,0}   -- DOS DOESN'T USE ANY_UP
end if
global constant -- Mouse Events:
        MOVE        = tmp[1]
       ,LEFT_DOWN   = tmp[2]
       ,LEFT_UP     = tmp[3]
       ,RIGHT_DOWN  = tmp[4]
       ,RIGHT_UP    = tmp[5]
       ,MIDDLE_DOWN = tmp[6]
       ,MIDDLE_UP   = tmp[7]
       ,ANY_UP      = tmp[8]

global function get_mouse() -- M_GET_MOUSE       = 14    --
-- report mouse events,
-- returns -1 if no mouse event,
-- otherwise returns {event#, x-coord, y-coord}
  return machine_func(M_GET_MOUSE, NULL)
end function

global procedure mouse_events(integer events) -- M_MOUSE_EVENTS    = 15    --
-- select the mouse events to be reported by get_mouse()
-- e.g. mouse_events(LEFT_UP + LEFT_DOWN + RIGHT_DOWN)
  machine_proc(M_MOUSE_EVENTS, events)
end procedure

global procedure mouse_pointer(boolean show_it) -- M_MOUSE_POINTER   = 24    --
-- show (1) or hide (0) the mouse pointer
  machine_proc(M_MOUSE_POINTER, show_it)
end procedure
