include std/core.em

-----------------------------------------------------------------------------
-- BITWISE MANIPULATORS --

------------------------------------------------
global function lo_byte( integer i1 )
------------------------------------------------
-- syntax: i2 = lo_byte( i1 )
-- description:
--    returns the low byte i2, from the word(2 bytes) i1

    -- return low byte
    return and_bits( i1, #FF )
end function

------------------------------------------------
global function hi_byte( integer i1 )
------------------------------------------------
-- syntax: i2 = hi_byte( i1 )
-- description:
--    returns the high byte i2, from the word(2 bytes) i1

    -- return high byte
    return floor( and_bits(i1,#FF00) / #100 )
end function

------------------------------------------------
global function lo_word( atom a )
------------------------------------------------
-- syntax: i = lo_word( a )
-- description:
--    returns the low word(2 bytes) i, from the long(4 bytes) a

    -- return low word
    return and_bits( a, #FFFF )
end function

------------------------------------------------
global function hi_word( atom a )
------------------------------------------------
-- syntax: i =  hi_word( a )
-- description:
--    returns the high word(2 bytes) i, from the long(4 bytes) a

    -- return high word
    return and_bits(floor( a / #10000 ), #FFFF)
end function

------------------------------------------------
global function make_word( integer lo, integer hi)
------------------------------------------------
-- syntax: i = make_word( lo, hi )
-- description:
--    make 2 bytes (8bit) into a word (16bit)

   -- mask values ( copied from Derek Parnell's tk_math.e )
   lo  = and_bits( lo,  #FF )
   hi = and_bits( hi, #FF )
   
   return hi*#100 + lo
end function

------------------------------------------------
global function make_long( integer lo, integer hi)
------------------------------------------------
-- syntax: a = make_long( lo, hi )
-- description:
--    make 2 words (16bit) into a long (32bit)

   -- mask values ( copied from Derek Parnell's tk_math.e )
   lo  = and_bits( lo,  #FFFF )
   hi = and_bits( hi, #FFFF )

   return hi*#10000 + lo
end function
------------------------------------------------


---------------------------------------------------------------------------------
global function int_to_bits(atom x, integer nbits)
--    Returns the low-order nbits bits of x as a sequence of 1's and 0's. 
--    Note that the least significant bits come first. You can use Euphoria's
--    and/or/not operators on sequences of bits. You can also subscript, 
--    slice, concatenate etc. to manipulate bits.
 sequence bits
 integer mask
    
  bits = repeat(0, nbits)
  if integer(x) and nbits < 30 then
  	-- faster method
	  mask = 1
	  for i = 1 to nbits do
	    bits[i] = and_bits(x, mask) and 1
	    mask *= 2
	  end for
  else
	  -- slower, but works for large x and large nbits
	  if x < 0 then
	    x += power(2, nbits) -- for 2's complement bit pattern
	  end if
	  for i = 1 to nbits do
	    bits[i] = remainder(x, 2) 
	    x = floor(x/2)
	  end for
  end if
  return bits
end function


global function bits_to_int(sequence bits)
-- get the (positive) value of a sequence of "bits"
 atom value, p
  value = 0
  p = 1
  for i = 1 to length(bits) do
  	if bits[i] then
	    value += p
	  end if
	  p += p
  end for
  return value
end function
---------------------------------------------------------------------------------


---------------------------------------------------------------------------------
constant mem = machine_func(M_ALLOC,4)

global function int_to_bytes(atom x)
-- returns value of x as a sequence of 4 bytes 
-- that you can poke into memory 
--      {bits 0-7,  (least significant)
--       bits 8-15,
--       bits 16-23,
--       bits 24-31} (most significant)
-- This is the order of bytes in memory on 386+ machines.
  poke4(mem,x)
  return peek({mem,4})
end function


global function bytes_to_int(sequence s)
-- converts 4-byte peek() sequence into an integer value
  poke(mem,s[1..4])
  return peek4u(mem)
end function
---------------------------------------------------------------------------------


---------------------------------------------------------------------------------
type sequence_8(sequence s)
-- an 8-element sequence
  return length(s) = 8
end type

type sequence_4(sequence s)
-- a 4-element sequence
  return length(s) = 4
end type

global function atom_to_float64(atom a)
-- Convert an atom to a sequence of 8 bytes in IEEE 64-bit format
    return machine_func(M_A_TO_F64, a)
end function

global function atom_to_float32(atom a)
-- Convert an atom to a sequence of 4 bytes in IEEE 32-bit format
    return machine_func(M_A_TO_F32, a)
end function

global function float64_to_atom(sequence_8 ieee64)
-- Convert a sequence of 8 bytes in IEEE 64-bit format to an atom
    return machine_func(M_F64_TO_A, ieee64)
end function

global function float32_to_atom(sequence_4 ieee32)
-- Convert a sequence of 4 bytes in IEEE 32-bit format to an atom
    return machine_func(M_F32_TO_A, ieee32)
end function
---------------------------------------------------------------------------------


------------------------------------------------
global function or_all(sequence s)
------------------------------------------------
-- syntax: a = or_all( s )
-- description:
--    combine a sequence of atoms using or_bits(), and return the value

 atom a
   a = 0

   for i = 1 to length(s) do
      a = or_bits(a,s[i])
   end for
   return a
end function
------------------------------------------------
