include std/core.em

		----------------------------------------
		-- Machine Level Programming for 386+ --
		-- * DEBUGGING VERSION of machine.e * --
		----------------------------------------

-- How To Use This File:

-- 1. Copy safe.e into the same directory as your main .ex[w] file and 
--    rename safe.e as machine.e in that directory. (Do NOT simply 
--    include safe.e, or you may have naming conflicts.)

-- 2. If your program doesn't already include machine.e add:  
--           include machine.e  
--    to your main .ex[w] file at the top.

-- 3. If necessary, call register_block(address, length) to add additional
--    "external" blocks of memory to the safe_address_list. These are blocks 
--    of memory that are safe to read/write but which you did not acquire 
--    through Euphoria's allocate() or allocate_low(). Call 
--    unregister_block(address) when you want to prevent further access to 
--    an external block.

-- 4. Run your program. It might be 10x slower than normal but it's
--    worth it to catch a nasty bug.

-- 5. If a bug is caught, you will hear some "beep" sounds.
--    Press Enter to clear the screen and see the error message. 
--    There will be a "divide by zero" traceback in ex.err 
--    (if you are eligible for diagnostic dumps) so you can find the
--    statement that is making the illegal memory access.

-- 6. When you are finished debugging and want to run at full speed,
--    remove or rename the local copy of machine.e (really safe.e) 
--    in your directory.

-- This file is equivalent to machine.e, but it overrides the built-in 
-- routines: 
--     poke, peek, poke4, peek4s, peek4u, call, mem_copy, and mem_set
-- and it provides alternate versions of:
--     allocate, allocate_low, free, free_low

integer all_calls, edges_only
all_calls = 1   -- check all blocks for edge corruption after each 
		  -- call(), dos_interrupt(), c_proc(), or c_func(). 
		  -- To save time, your program can turn off this checking by 
		  -- setting all_calls to 0. 

edges_only = 0    -- only check for references to the leader or trailer
		  -- areas just outside each registered block.
		  -- don't complain about addresses that are far out of bounds
		  -- (it's probably a legitimate block from another source)
		  -- Set this to 1 for Win32Lib-based programs.
		  
-- Your program will only be allowed to read/write areas of memory
-- that it allocated (and hasn't freed), as well as areas in low memory 
-- that you list below, or add dynamically via register_block().

sequence safe_address_list
-- Include the starting address and length of any 
-- acceptable areas of memory for peek/poke here. 
-- Set allocation number to 0.
if platform() = 1 then -- DOS32
    safe_address_list = {
--      {start , length , allocation_number}        
	{#A0000, 200*320, 0},   -- mode 19 pixel memory, start & length 
      --{#B0000, 4000   , 0},   -- monochrome text memory, first page
	{#B8000, 8000   , 0},   -- color text memory, first page, 50-line mode 
	{1024  , 100    , 0}    -- keyboard buffer area (roughly)
	-- add more here
}
else
  safe_address_list = {}
end if

constant machine_func_rID = callback_id("machine_func")
constant machine_proc_rID = callback_id("machine_proc")
constant poke_rID     = callback_id("poke")
constant peek_rID     = callback_id("peek")
constant poke4_rID    = callback_id("poke4")
constant peek4s_rID   = callback_id("peek4s")
constant peek4u_rID   = callback_id("peek4u")
constant mem_copy_rID = callback_id("mem_copy")
constant mem_set_rID  = callback_id("mem_set")
constant call_rID     = callback_id("call")
constant c_proc_rID   = callback_id("c_proc")
constant c_func_rID   = callback_id("c_func")


---------------------------------------------------------------------------------
with type_check
puts(1, "\n\t\tUsing Debug Version of machine.e\n")
atom t    t = time()
while (time()-t) < 3 do end while

constant BORDER_SPACE = 40
constant leader = repeat('@', BORDER_SPACE)
constant trailer = repeat('%', BORDER_SPACE)


procedure die(sequence msg)
-- Terminate with a message.
-- makes warning beeps first so you can see what's happening on the screen
  beep(1000, 0.1, 0.1, 7)
  any_key("\n *** Press Enter *** ")
  if graphics_mode(-1) then end if -- restore default graphics_mode
  fatal_error("\n\n" & msg & "\n\n")
end procedure
define_callback("error_routine",routine_id("die")) -- error handler


function bad_address(atom a)
-- show address in decimal and hex  
  return sprintf(" ADDRESS!!!! %d (#%08x)", {a, a})
end function


function safe_address(atom start, integer len)
-- is it ok to read/write all addresses from start to start+len-1?
 atom block_start, block_upper, upper
 sequence block
    
  upper = start + len
  -- search the list of safe memory blocks:
  for i = 1 to length(safe_address_list) do
  	block = safe_address_list[i]
	  block_start = block[1]
	  if edges_only then
	    -- addresses are considered safe as long as 
	    -- they aren't in any block's border zone
	    if start <= 3 then
		    return FALSE -- null pointer (or very small address)
	    end if
	    if block[3] >= 1 then
		    -- an allocated block with a border area
		    block_upper = block_start + block[2]
		    if (start >= block_start - BORDER_SPACE and 
		        start < block_start) or 
		       (start >= block_upper and 
		        start < block_upper + BORDER_SPACE) then
		      return FALSE
		    elsif (upper > block_start - BORDER_SPACE and
		        upper <= block_start) or
		       (upper > block_upper and
		        upper < block_upper + BORDER_SPACE) then
		      return FALSE
		    elsif start < block_start - BORDER_SPACE and
		        upper > block_upper + BORDER_SPACE then
		      return FALSE
		    end if
	    end if
	  else
  	  -- addresses are considered safe as long as 
  	  -- they are inside an allocated or registered block
  	  if start >= block_start then 
  		  block_upper = block_start + block[2]
  		  if upper <= block_upper then
  		    if i > 1 then
  			    -- move block i to the top and move 1..i-1 down
  			    if i = 2 then
  			      -- common case, subscript is faster than slice:
  			      safe_address_list[2] = safe_address_list[1]
  			    else
  			      safe_address_list[2..i] = safe_address_list[1..i-1]
  			    end if
  			    safe_address_list[1] = block
  		    end if
  		    return TRUE
  		  end if
  	  end if
  	end if
  end for
  if not edges_only then return FALSE end if -- not found in any safe block
	return TRUE  -- not found in any border zone
end function


global procedure check_calls(boolean yes_or_no)
  all_calls = yes_or_no
end procedure

global procedure check_edges_only(boolean yes_or_no)
  edges_only = yes_or_no
end procedure

global procedure register_block(machine_addr block_addr, positive_int block_len)
-- register an externally-acquired block of memory as being safe to use
  allocation_num += 1
  safe_address_list = prepend(safe_address_list
                             ,{block_addr, block_len, -allocation_num})
end procedure

global procedure unregister_block(machine_addr block_addr)
-- remove an external block of memory from the safe address list
  for i = 1 to length(safe_address_list) do
  	if safe_address_list[i][1] = block_addr then
	    if safe_address_list[i][3] >= 0 then
	    	error("ATTEMPT TO UNREGISTER A NON-EXTERNAL BLOCK")
	    end if
	    safe_address_list = safe_address_list[1..i-1] &
			       safe_address_list[i+1..length(safe_address_list)]
	   return
	  end if  
  end for
  error("ATTEMPT TO UNREGISTER A BLOCK THAT WAS NOT REGISTERED!")
end procedure


function prepare_block(atom a, integer n)
-- set up an allocated block so we can check it for corruption
  if a = NULL then error("OUT OF MEMORY!") end if
  call_proc(poke_rID,{a, leader})
  a += BORDER_SPACE
  call_proc(poke_rID,{a+n, trailer})
  allocation_num += 1
--  if allocation_num = ??? then 
--      trace(1) -- find out who allocated this block number
--  end if  
  safe_address_list = prepend(safe_address_list, {a, n, allocation_num})
  return a
end function


atom allocation_num
allocation_num = 0

procedure show_byte(atom m)
-- display byte at memory location m
 integer c
    
  c = call_func(peek_rID,{m})
  if c <= 9 then
  	printf(1, "%d", c)
  elsif c < 32 or c > 127 then
	  printf(1, "%d #%02x", {c, c})
  else
	  if c = leader[1] or c = trailer[1] then
	    printf(1, "%s", c)
	  else
	    printf(1, "%d #%02x '%s'", {c, c, c})
	  end if
  end if
  puts(1,",  ")
end procedure

procedure show_block(sequence block_info)
-- display a corrupted block and die
 integer len, id, bad, p
 atom start
    
  start = block_info[1]
  len = block_info[2]
  id = block_info[3]
  printf(1, "BLOCK# %d, START: #%x, SIZE %d\n", {id, start, len})
  -- check pre-block
  bad = 0
  for i = start-BORDER_SPACE to start-1 do
  	p = call_func(peek_rID,{i})
	  if p != leader[1] or bad then
	    bad += 1
	    if bad = 1 then
		    puts(1, "DATA WAS STORED ILLEGALLY, JUST BEFORE THIS BLOCK:\n")
		    puts(1, "(" & leader[1] & " characters are OK)\n")
		    printf(1, "#%x: ", i)
	    end if
	    show_byte(i)
	  end if
  end for
  puts(1, "\nDATA WITHIN THE BLOCK:\n")
  printf(1, "#%x: ", start)
  if len <= 30 then
	  -- show whole block
	  for i = start to start+len-1 do
	    show_byte(i)
	  end for 
  else
	  -- first part of block
	  for i = start to start+14 do
	    show_byte(i)
	  end for 
	  -- last part of block
	  puts(1, "\n ...\n")
	  printf(1, "#%x: ", start+len-15)
	  for i = start+len-15 to start+len-1 do
	    show_byte(i)
	  end for 
  end if
  bad = 0
  -- check post-block
  for i = start+len to start+len+BORDER_SPACE-1 do
	  p = call_func(peek_rID,{i})
	  if p != trailer[1] or bad then
	    bad += 1
	    if bad = 1 then
		    puts(1, "\nDATA WAS STORED ILLEGALLY, JUST AFTER THIS BLOCK:\n")
		    puts(1, "(" & trailer[1] & " characters are OK)\n")
		    printf(1, "#%x: ", i)
	    end if
	    show_byte(i)
	  end if
  end for 
  error("BLOCK IS CORRUPT!")
end procedure

global procedure check_all_blocks()
-- Check all allocated blocks for corruption of the leader and trailer areas. 
 integer n
 atom a
 sequence block

  if all_calls then
    for i = 1 to length(safe_address_list) do
    	block = safe_address_list[i]
  	  if block[3] >= 1 then
  	    -- a block that we allocated
  	    a = block[1]
  	    n = block[2]
  	    if not equal(leader, call_func(peek_rID,{{a-BORDER_SPACE, BORDER_SPACE}})) then
  		    show_block(block)
  	    elsif not equal(trailer, call_func(peek_rID,{{a+n, BORDER_SPACE}})) then
  		    show_block(block)
  	    end if          
  	  end if
    end for
	end if
end procedure

---------------------------------------------------------------------------------
without warning
global procedure poke(atom a, object v)
-- safe version of poke 
 integer len
    
  if atom(v) then
  	len = 1
  else
	  len = length(v)
  end if
  if safe_address(a, len) then
	  call_proc(poke_rID,{a,v})
  else
	  die("BAD POKE" & bad_address(a))
  end if
end procedure
define_callback("poke",routine_id("poke"))

-- override "peek" with debug peek
global function peek(object x)
-- safe version of peek 
 integer len
 atom a
    
  if atom(x) then
  	len = 1
	  a = x
  else
	  len = x[2]
	  a = x[1]
  end if
  if safe_address(a, len) then
	  return call_func(peek_rID,{x})
  else
	  die("BAD PEEK" & bad_address(a))
  end if
end function
define_callback("peek",routine_id("peek"))


global procedure poke4(atom a, object v)
-- safe version of poke4 
 integer len
    
  if atom(v) then
  	len = 4
  else
	  len = length(v)*4
  end if
  if safe_address(a, len) then
	  call_proc(poke4_rID,{a,v})
  else
	  die("BAD POKE4" & bad_address(a))
  end if
end procedure
define_callback("poke4",routine_id("poke4"))

-- override "peek4s" with debug peek4s
global function peek4s(object x)
-- safe version of peek4s 
 integer len
 atom a

  if atom(x) then
  	len = 4
	  a = x
  else
	  len = x[2]*4
	  a = x[1]
  end if
  if safe_address(a, len) then
	  return call_func(peek4s_rID,{x})
  else
	  die("BAD PEEK4S" & bad_address(a))
  end if
end function
define_callback("peek4s",routine_id("peek4s"))


-- override "peek4u" with debug peek4u
global function peek4u(object x)
-- safe version of peek4u 
 integer len
 atom a
    
  if atom(x) then
  	len = 4
	  a = x
  else
	  len = x[2]*4
	  a = x[1]
  end if
  if safe_address(a, len) then
	  return call_func(peek4u_rID,{x})
  else
	  die("BAD PEEK4U" & bad_address(a))
  end if
end function
define_callback("peek4u",routine_id("peek4u"))


global procedure mem_copy(machine_addr target, machine_addr source, natural len)
-- safe mem_copy
  if not safe_address(target, len) then 
	  die("BAD MEM_COPY TARGET" & bad_address(target))
  elsif not safe_address(source, len) then
	  die("BAD MEM_COPY SOURCE" & bad_address(source))
  else
	  call_proc(mem_copy_rID,{target, source, len})
  end if
end procedure
define_callback("mem_copy",routine_id("mem_copy"))


global procedure mem_set(machine_addr target, atom value, natural len)
-- safe mem_set
  if safe_address(target, len) then
  	call_proc(mem_set_rID,{target, value, len})
  else
	  die("BAD MEM_SET" & bad_address(target))
  end if
end procedure
define_callback("mem_set",routine_id("mem_set"))


global procedure call(atom addr)
-- safe call - machine code must start in block that we own
  if safe_address(addr, 1) then
  	call_proc(call_rID,{addr})
    check_all_blocks() -- check for any corruption
  else
	  die(sprintf("BAD CALL ADDRESS!!!! %d\n\n", addr))
  end if
end procedure
define_callback("call",routine_id("call"))


global procedure c_proc(integer i, sequence s)
  call_proc(c_proc_rID,{i,s})
	check_all_blocks()
end procedure
define_callback("c_proc",routine_id("c_proc"))


global function c_func(integer i, sequence s)
 object r
  r = call_func(c_func_rID,{i, s})
 	check_all_blocks()
  return r
end function
define_callback("c_func",routine_id("c_func"))


---------------------------------------------------------------------------------
-- SAFE MACHINE ROUTINES --

function safe_allocate(positive_int n)
-- allocate memory block and add it to safe list
 atom a
 integer half
  a = call_func(machine_func_rID,{M_ALLOC, n+BORDER_SPACE*2})
  return prepare_block(a, n)
end function

function safe_allocate_low(positive_int n)
-- allocate memory block and add it to safe list
 atom a
  a = call_func(machine_func_rID,{M_ALLOC_LOW, n+BORDER_SPACE*2})
  return prepare_block(a, n)
end function

procedure safe_free(machine_addr a)
-- free address a - make sure it was allocated
 integer n
  for i = 1 to length(safe_address_list) do
  	if safe_address_list[i][1] = a then
	    -- check pre and post block areas
	    if safe_address_list[i][3] <= 0 then
	    	die("ATTEMPT TO FREE A BLOCK THAT WAS NOT ALLOCATED!")
	    end if
	    n = safe_address_list[i][2]
	    if not equal(leader, call_func(peek_rID,{{a-BORDER_SPACE, BORDER_SPACE}})) then
		    show_block(safe_address_list[i])
	    elsif not equal(trailer, call_func(peek_rID,{{a+n, BORDER_SPACE}})) then
		    show_block(safe_address_list[i])
	    end if          
	    call_proc(machine_proc_rID,{M_FREE, a-BORDER_SPACE})
	    -- remove it from list
	    safe_address_list = safe_address_list[1..i-1]
	        & safe_address_list[i+1..length(safe_address_list)]
	    return
	  end if
  end for
  die("ATTEMPT TO FREE USING AN ILLEGAL ADDRESS!")
end procedure

procedure safe_free_low(low_machine_addr a)
-- free low address a - make sure it was allocated
 integer n
  if a > 1024*1024 then
  	die("TRYING TO FREE A HIGH ADDRESS USING free_low!")
  end if
  for i = 1 to length(safe_address_list) do
	  if safe_address_list[i][1] = a then
	    -- check pre and post block areas
	    if safe_address_list[i][3] <= 0 then
		    die("ATTEMPT TO FREE A BLOCK THAT WAS NOT ALLOCATED!")
	    end if
	    n = safe_address_list[i][2]
	    if not equal(leader, call_func(peek_rID,{{a-BORDER_SPACE, BORDER_SPACE}})) then
		    show_block(safe_address_list[i])
	    elsif not equal(trailer, call_func(peek_rID,{{a+n, BORDER_SPACE}})) then
		    show_block(safe_address_list[i])
	    end if          
	    call_proc(machine_proc_rID,{M_FREE_LOW, a-BORDER_SPACE})
	    -- remove it from list
	    safe_address_list = 
			safe_address_list[1..i-1] &
			safe_address_list[i+1..length(safe_address_list)]
	    return
	  end if
  end for
  die("ATTEMPT TO FREE USING AN ILLEGAL ADDRESS!")
end procedure


function safe_dos_interrupt(integer int_num, register_list input_regs)
-- call the DOS operating system via software interrupt int_num, using the
-- register values in input_regs. A similar register_list is returned.
-- It contains the register values after the interrupt.
 object r
  r = call_func(machine_func_rID,{M_INTERRUPT, {int_num, input_regs}})
  check_all_blocks()
  return r
end function


global function machine_func(integer mID, object args)
  if    mID = M_ALLOC then
    return safe_allocate(args)
  elsif mID = M_ALLOC_LOW then
    return safe_allocate_low(args)
  elsif mID = M_INTERRUPT then
    return safe_dos_interrupt(args[1],args[2])
  else
    return call_func(machine_func_rID,{mID,args})
  end if
end function
define_callback("machine_func",routine_id("machine_func"))

global procedure machine_proc(integer mID, object args)
  if    mID = M_FREE then
    safe_free(args)
  elsif mID = M_FREE_LOW then
    safe_free_low(args)
  else
    call_proc(machine_proc_rID,{mID,args})
  end if
end procedure
define_callback("machine_proc",routine_id("machine_proc"))

