include std/core.em

---------------------------------------------------------------------------------
-- TYPES AND DEFINITIONS --

-- biggest address on a 32-bit machine
constant MAX_ADDR = power(2, 32)-1
-- biggest address accessible to 16-bit real mode
constant LOW_ADDR = power(2, 20)-1


global type pointer(object addr)
-- a 32-bit non-null machine address 
 return atom(addr) and (floor(addr) = addr) -- integer is only 31 bit, pointers are 32
    and (1 <= addr) and (addr <= MAX_ADDR)
end type

global type near_pointer(object addr)
-- a legal low machine address 
 return integer(addr)
    and (1 <= addr) and (addr <= LOW_ADDR)
end type

global type far_pointer(sequence addr)
-- protected mode far address {seg, offset}
 return length(addr)=2
    and integer(addr[1])
    and pointer(addr[2])
end type

global constant
        REG_DI    = 1
       ,REG_SI    = 2
       ,REG_BP    = 3
       ,REG_BX    = 4
       ,REG_DX    = 5
       ,REG_CX    = 6
       ,REG_AX    = 7
       ,REG_FLAGS = 8 -- on input: ignored, on output: low bit has carry flag for success/fail
       ,REG_ES    = 9
       ,REG_DS    = 10
global constant REG_LIST_SIZE = 10

global type register_list(sequence r)
-- a list of register values
    return length(r) = REG_LIST_SIZE
end type


---------------------------------------------------------------------------------
-- ALLOCATION ROUTINES --

global function alloc(positive_int n)
-- Allocate n bytes of memory and return the address.
-- Free the memory using free() below.
 return machine_func(M_ALLOC, n)
end function

global procedure dealloc(pointer addr)
-- free the memory at addr
  machine_proc(M_FREE, addr)
end procedure

global function alloc_sz(sequence s)
-- create a C-style null-terminated string in memory
 atom mem
  mem = alloc(length(s) + 1)
  if mem then
    poke(mem, s)
	  poke(mem+length(s), NULL)
  end if
 return mem
end function

---------------------------------------------------------------------------------
-- LOW MEMORY ALLOCATION ROUTINES --

global function alloc_low(positive_int n)
-- Allocate n bytes of low memory (address less than 1Mb) 
-- and return the address. Free this memory using free_low() below.
-- Addresses in this range can be passed to DOS during software interrupts.
    return machine_func(M_ALLOC_LOW, n)
end function

global procedure dealloc_low(near_pointer addr)
-- free the low memory at addr
    machine_proc(M_FREE_LOW, addr)
end procedure

global procedure lock_memory(pointer addr, positive_int n)
-- Prevent a chunk of code or data from ever being swapped out to disk.
-- You should lock any code or data used by an interrupt handler.
    machine_proc(M_LOCK_MEMORY, {addr, n})
end procedure

---------------------------------------------------------------------------------
-- DOS INTERRUPT ROUTINES --

global function dos_interrupt(integer int_num, register_list input_regs)
-- call the DOS operating system via software interrupt int_num, using the
-- register values in input_regs. A similar register_list is returned.
-- It contains the register values after the interrupt.
    return machine_func(M_INTERRUPT, {int_num, input_regs})
end function

global function get_vector(integer int_num)
-- returns the current (far) address of the interrupt handler
-- for interrupt vector number int_num as a 2-element sequence: 
-- {16-bit segment, 32-bit offset}
    return machine_func(M_GET_VECTOR, int_num)
end function

global procedure set_vector(integer int_num, far_pointer a)
-- sets a new interrupt handler address for vector int_num  
    machine_proc(M_SET_VECTOR, {int_num, a})
end procedure


---------------------------------------------------------------------------------
-- MEMORY DATA MANIPULATORS --

type pointer_object(object x)
  if atom(x) then
   return pointer(x)
  else
   return length(x)=2
      and pointer(x[1])
      and positive_int(x[2])
  end if
end type

------------------------------------------------
global procedure poke_string(pointer addr, sequence s)
------------------------------------------------
-- syntax: poke_string( a, s )
-- description:
--    pokes the string sequence s, to address a, as a null-terminated string
   poke(addr, s)
   poke(addr+length(s), 0)
end procedure

------------------------------------------------
global function peek_string(pointer addr)
------------------------------------------------
-- syntax: s = peek_string( x )
-- description:
--    retrieves a null-terminated string from address x.
--    if x is a sequence, it is assumed to be an array of addresses,
--       an array of strings will be returned respectively
 atom offset
  offset = addr
  while peek(offset) do offset +=1 end while
  return peek({addr,offset-addr})
end function

------------------------------------------------
global procedure poke2(pointer addr, object x)
------------------------------------------------
-- syntax: poke2( a, x )
-- description:
--    pokes x as a word, to addr
--    if x is a sequence, it will be poked as an array of words
  if atom(x) then x = {x} end if

  for i = 1 to length(x) do
    poke(addr,{and_bits( x[i], #FF ), floor( x[i]/#100 )})
    addr += 2
  end for
end procedure

------------------------------------------------
global function peek2u(pointer_object x)
------------------------------------------------
-- syntax: i = peek2u( x )
-- description:
--    returns a 2 byte unsigned word from address x
--    if x is a sequence, it is expected to be of the form: {addr, len}
--       where addr is the base address of the array, and len is the number of words to retrieve
 sequence ret
 
  if atom(x) then
    return peek(x+1)*#100 + peek(x)
  else
    ret = repeat(0,x[2])
    x = x[1]
    for i = 1 to length(ret) do
      ret[i] = peek(x+1)*#100 + peek(x)
      x += 2
    end for
    return ret
  end if
end function

------------------------------------------------
global function peek2s(pointer_object x)
------------------------------------------------
-- syntax: i = peek2s( x )
-- description:
--    returns a 2 byte signed word from address x
--    if x is a sequence, it is expected to be of the form: {addr, len}
--       where addr is the base address of the array, and len is the number of words to retrieve
 object ret
   if atom(x) then
      ret = peek(x+1)*#100 + peek(x)
      if ret >= #8000 then ret -= #10000 end if      
      return ret
   else
      ret = repeat(0,x[2])
      x = x[1]
      for i = 1 to length(ret) do
         ret[i] = peek(x+1)*#100 + peek(x)
         if ret[i] >= #8000 then ret[i] -= #10000 end if
         x += 2
      end for
      return ret
   end if
end function

------------------------------------------------
global procedure poke_float32(pointer addr, object x)
------------------------------------------------
-- syntax: poke_float32( a, x )
-- description:
--    pokes x as a 32bit float, to address a
--    if x is a sequence, it will be poked as an array of 32bit floats

   if atom(x) then x = {x} end if

   for i = 1 to length(x) do
      poke(addr, machine_func(M_A_TO_F32, x[i]))
      addr += 4
   end for   
end procedure

------------------------------------------------
global function peek_float32(pointer_object x)
------------------------------------------------
-- syntax: a = peek_float32( x )
-- description:
--    returns a 32bit float value from address x
--    if x is a sequence, it is expected to be of the form: {addr, len}
--       where addr is the base address of the array, and len is the number of 32bit floats to retrieve
 object ret
 sequence bytes
 integer a
   if atom(x) then
      return machine_func(M_F32_TO_A,peek({x, 4}))
   else
      ret = repeat(0,x[2])
      bytes = peek({x[1], x[2]*4})
      a = 1
      for i = 1 to length(ret) do
         ret[i] = machine_func(M_F32_TO_A,bytes[a..a+3])
         a += 4
      end for
      return ret
   end if
end function

------------------------------------------------
global procedure poke_float64(pointer addr, object x)
------------------------------------------------
-- syntax: poke_float64( a, x )
-- description:
--    pokes x as a 64bit float, to address a
--    if x is a sequence, it will be poked as an array of 64bit float's

   if atom(x) then x = {x} end if
   
   for i = 1 to length(x) do
      poke(addr, machine_func(M_A_TO_F64, x[i]))
      addr += 8
   end for
end procedure

------------------------------------------------
global function peek_float64(pointer_object x)
------------------------------------------------
-- syntax: a = peek_float64( x )
-- description:
--    returns a 64bit float value from address x
--    if x is a sequence, it is expected to be of the form: {addr, len}
--       where addr is the base address of the array, and len is the number of 64bit floats to retrieve
 object ret
 sequence bytes
 integer a
   if atom(x) then      
      return machine_func(M_F64_TO_A,peek({x, 8}))
   else
      ret = repeat(0,x[2])
      bytes = peek({x[1], x[2]*8})
      a = 1
      for i = 1 to length(ret) do
         ret[i] = machine_func(M_F64_TO_A,bytes[a..a+7])
         a += 8
      end for
      return ret
   end if
end function


---------------------------------------------------------------------------------
-- DEBUG ROUTINE PLACEHOLDERS --

-- variables and routines used in safe.e
without warning

global procedure check_calls(atom yes_or_no)
end procedure

global procedure check_edges_only(atom yes_or_no)
end procedure

global procedure register_block(atom block_addr, atom block_len)
end procedure

global procedure unregister_block(atom block_addr)
end procedure

global procedure check_all_blocks()
end procedure
with warning
