include std/core.em

global constant LOCK_SHARED     = 1
global constant LOCK_EXCLUSIVE  = 2

global type file_num(integer f)
  return f >= -1
end type

global type file_pos(atom p)
  return p >= -1
end type

-------------------------------------------------
-- COMMON --
global procedure seek(file_num fn, file_pos pos)
-- Seeks to a byte position in the file, 
-- or to end of file if pos is -1.
-- This function is normally used with
-- files opened in binary mode.
  machine_proc(M_SEEK, {fn, pos})
end procedure

global function safe_seek(file_num fn, file_pos pos)
-- This is the old, RDS seek() routine
  return machine_func(M_SEEK, {fn, pos})
end function

global function where(file_num fn)
-- Returns the current byte position in the file.
-- This function is normally used with
-- files opened in binary mode.
    return machine_func(M_WHERE, fn)
end function

global function file_size(file_num fn)
-- seek to EOF and return the byte position in the file (the size)
-- this routine does NOT restore the original byte position
  seek(fn,-1)
  return where(fn)
end function

global procedure flush(file_num fn)
-- flush out the buffer associated with file fn
    machine_proc(M_FLUSH, fn)
end procedure


-------------------------------------------------
-- LOCK ROUTINES --
type lock_type(integer t)
  if platform() = LINUX then
  	return t = LOCK_SHARED or t = LOCK_EXCLUSIVE
  else
	  return 1
  end if
end type

type byte_range(sequence r)
  if length(r) = 0 then
	  return 1
  elsif length(r) = 2 and r[1] <= r[2] then
	  return 1
  else
	  return 0
  end if
end type

global function lock_file(file_num fn, lock_type t, byte_range r)
-- Attempt to lock a file so other processes won't interfere with it.
-- The byte range can be {} if you want to lock the whole file
    return machine_func(M_LOCK_FILE, {fn, t, r})
end function

global procedure unlock_file(file_num fn, byte_range r) 
-- The byte range can be {} if you want to unlock the whole file.
    machine_proc(M_UNLOCK_FILE, {fn, r})
end procedure


---------------------------------------------------------------------------------
-- DIR ROUTINES --
global constant 
	D_NAME        = 1
 ,D_ATTRIBUTES  = 2
 ,D_SIZE        = 3
 ,D_YEAR        = 4
 ,D_MONTH       = 5
 ,D_DAY         = 6
 ,D_HOUR        = 7
 ,D_MINUTE      = 8
 ,D_SECOND      = 9

------------------------------------------------
global function current_dir()
-- returns name of current working directory
  return machine_func(M_CURRENT_DIR, 0)
end function

global function chdir(sequence newdir)
-- Changes the current directory. Returns 1 - success, 0 - fail.
  return machine_func(M_CHDIR, newdir)
end function

global function dir(sequence name)
-- returns directory information, given the name
-- of a file or directory. Format returned is:
-- {
--  {"name1", attributes, size, year, month, day, hour, minute, second},
--  {"name2", ...                                                     },
-- }
 object d
  d = machine_func(M_DIR, name)
  if atom(d) then return {} end if
  return d
end function

global function dir_trimmed(sequence d)
-- syntax:        d = trim_dlist(d)
-- returns: the dir list, without "." or ".." entries
-- description:
--    trims the "." and ".." entries from a dir sequence
  if length(d) then
    if atom(d[1]) then d = dir(d) end if -- d is not a dir list
    if equal(d[1][1],".") then
      if length(d) > 1 then
        if equal(d[2][1],"..") then return d[3..length(d)] end if
      end if
      return d[2..length(d)]
    end if
  end if
  return d
end function

global function dir_tree(string path)
-- syntax:        
-- returns: a dir listing for path, including subdirectories
-- description:
--    the dir listing is trimmed (no "." or ".." entries)
--    and path is prepended to the name field for each entry
 sequence d, subd, subname
  d = dir_trimmed(path)
  for i = 1 to length(d) do
    if find('d',d[i][D_ATTRIBUTES]) then
      subname = d[i][D_NAME]
      subd = dir_tree(path&SLASH&subname)
      for j = 1 to length(subd) do
        subd[j][D_NAME] = subname&SLASH&subd[j][D_NAME]
      end for
      d &= subd
    end if
  end for
  return d
end function


---------------------------------------------------------------------------------
-- FILE IO ROUTINES --
-- written by Jason Mirwald
-- NOTE: max number of files that can be open at one time
-- by Euphoria is 25 including 0,1,2 (in/out/error)

-------------------------------------------------
global function getn(integer fn, integer n)
-- Return a sequence of n bytes from a file.
-- If EOF is reached before n bytes have been read, the remainder will be filled with -1's
 sequence s
  s = repeat(-1, n)
  for i = 1 to n do
    s[i] = getc(fn)
  end for
  return s
end function

------------------------------------------------
global function get_string(integer fn)
------------------------------------------------
-- syntax:        s = get_string(fn)
-- returns: a string of chars, up to, but not including binary 0 or EOF
-- description:
--    Get a 0-terminated string from a file.
------------------------------------------------
 sequence s
 integer c
  s = {}
  c = getc(fn)
  while c > 0 do
    s &= c
    c = getc(fn)    
  end while
  return s
end function
------------------------------------------------

----------------------------------------------

global function get2u(object o)
 sequence s
 integer fn
  if atom(o) then
    return getc(o) + (getc(o)*#0100)
  else
    if length(o) = 2 then
      fn = o[1]
      s = repeat(0,o[2])
      for n = 1 to o[2] do
        s[n] = getc(fn) + (getc(fn)*#0100)
      end for
      return s
    end if
  end if
  return 0
end function

----------------------------------------------

global function get2s(object o)
 sequence s
 integer fn
 atom a
  if atom(o) then
    a = getc(o) + (getc(o)*#0100)
    return a-(#10000*(and_bits(a,#8000)!=0))
  else
    if length(o) = 2 then
      fn = o[1]
      s = repeat(0,o[2])
      for n = 1 to o[2] do
        a = getc(fn) + (getc(fn)*#0100)
        s[n] = a-(#10000*(and_bits(a,#8000)!=0))
      end for
      return s
    end if
  end if
  return 0
end function

----------------------------------------------

global function get4u(object o)
 sequence s
 integer fn
  if atom(o) then
    return  getc(o)
          + getc(o)*#0100
          + getc(o)*#010000
          + getc(o)*#01000000
  else
    if length(o) = 2 then
      fn = o[1]
      s = repeat(0,o[2])
      for n = 1 to o[2] do
        s[n] =  getc(fn)
              + getc(fn)*#0100
              + getc(fn)*#010000
              + getc(fn)*#01000000
      end for
      return s
    end if
  end if
  return 0
end function

----------------------------------------------

global function get4s(object o)
 sequence s
 integer fn
 atom a
  if atom(o) then
    a = getc(o)
      + getc(o)*#0100
      + getc(o)*#010000
      + getc(o)*#01000000
    return a-(#100000000*( and_bits(a,#80000000)!=0 ))
  else
    if length(o) = 2 then
      fn = o[1]
      s = repeat(0,o[2])
      for n = 1 to o[2] do
        a = getc(fn)
          + getc(fn)*#0100
          + getc(fn)*#010000
          + getc(fn)*#01000000
        s[n] = a-(#100000000*( and_bits(a,#80000000)!=0 ))
      end for
      return s
    end if
  end if
  return 0
end function

----------------------------------------------

global function get_float32(object o)
 sequence s
 integer fn
  if atom(o) then
    return machine_func(49, getn(o,4))
  else
    if length(o) = 2 then
      fn = o[1]
      s = repeat(0,o[2])
      for n = 1 to o[2] do
        s[n] = machine_func(49, getn(fn,4))
      end for
      return s
    end if
  end if
  return 0
end function

----------------------------------------------

global function get_float64(object o)
 sequence s
 integer fn
  if atom(o) then
    return machine_func(47, getn(o,8))
  else
    if length(o) = 2 then
      fn = o[1]
      s = repeat(0,o[2])
      for n = 1 to o[2] do
        s[n] = machine_func(47, getn(fn,8))
      end for
      return s
    end if
  end if
  return -1
end function

----------------------------------------------

global procedure put2( integer fn, object o )
 sequence s
 atom a
 s = {}
 if atom(o) then
  o = {o}
 end if
 for n = 1 to length(o) do
  a = o[n]
  if a < 0 then
   a = #FFFF+a+1
  end if
  puts(fn,{remainder(a,#100),floor(a/#100)})
 end for
end procedure

----------------------------------------------

global procedure put4(integer fn, object o)
 atom a
 if atom(o) then
  o = {o}
 end if
 for n = 1 to length(o) do
  a = o[n]
  if a < 0 then
   a = #FFFFFFFF+a+1
  end if
  puts(fn,and_bits(#FF,{a,a/#100,a/#10000,a/#1000000}))
 end for
end procedure

----------------------------------------------

global procedure put_float32(integer fn, object o)
 if atom(o) then
  o = {o}
 end if
 for n = 1 to length(o) do
  puts(fn,machine_func(48,o[n]))
 end for
end procedure

----------------------------------------------

global procedure put_float64(integer fn, object o)
 if atom(o) then
  o = {o}
 end if
 for n = 1 to length(o) do
  puts(fn,machine_func(46,o[n]))
 end for
end procedure
