---------------------------------------------------------------------------------
-- vmm.e
-- version 1.1a
-- written by Chris Bensler      mailto: bensler@mail.com
---------------------------------------------------------------------------------
global constant VMM_VERSION = 1
---------------------------------------------------------------------------------
include std/core.e
include std/mem.e
include std/file.e


constant HEAP_STEP = 512

global constant VMM_ERROR_OPEN  = -1
global constant VMM_ERROR_INVALID = -2

global constant VMM_MAGIC_ID  = 0
global constant VMM_VERSION_NUMBER   = 3
global constant VMM_EXTRA_PTR = 4
global constant VMM_NUM_FREE  = 8
global constant VMM_FREE_PTR  = 12
global constant   VMM_HEADER_SIZE = 16


integer fnActive  fnActive = -1
integer nFree     nFree = 0
integer fpFree    fpFree = 0
integer sizeFree  sizeFree = 0

-- File Header --> 16 bytes
--          0   LONG        : Magic ID, always (VMM)
--          3   BYTE        : Version number in hex value: '0' is version 48
--          4   LONG        : Pointer to Extra Header Info
--          8   LONG        : Number of Free Blocks
--         12   LONG        : Pointer to Free Blocks List

-- Allocated Data Block --> 4 + align4(sizeof(data))
--   -4 | LONG        : Allocated size of block, starting from 0
--    0 | size        : Data, always 4 byte aligned

-- Free Blocks List --> 8*n bytes
-- The Free Blocks List is a vmm allocated data block
--    8*(n-1)   LONG        : Size of nth Entry
--  8*(n-1)+4   LONG        : Address of nth Entry

---------------------------------------------------------------------------------
-- UTILITY ROUTINES --
global procedure seek_vmm(integer x)
  machine_proc(M_SEEK,{fnActive,x})
end procedure

constant lp4 = alloc(4)
global function get4u_vmm()
  poke(lp4  ,getc(fnActive))
  poke(lp4+1,getc(fnActive))
  poke(lp4+2,getc(fnActive))
  poke(lp4+3,getc(fnActive))
  return peek4u(lp4)
end function

global procedure put4_vmm(integer x)
  poke4(lp4,x)
  puts(fnActive,peek({lp4,4}))
end procedure

global function getn_vmm(integer n)
 sequence x
  x = repeat(-1,n)
  for i = 1 to n do
    x[i] = getc(fnActive)
  end for
  return x
end function

global procedure putn_vmm(sequence x)
  puts(fnActive,x)
end procedure

procedure go_get_nFree()
  seek_vmm(VMM_NUM_FREE)
  nFree = get4u_vmm()
end procedure

procedure go_put_nFree(integer x)
  nFree = x
  seek_vmm(VMM_NUM_FREE)
  put4_vmm(nFree)
end procedure

constant ALIGN4 = {0,3,2,1}
global function align4(integer l)
  return l + ALIGN4[remainder(l,4)+1] 
end function

---------------------------------------------------------------------------------
-- FILE ROUTINES --
global function select_vmm(integer fn)
 integer lastActive
  lastActive = fnActive
  fnActive = fn
  seek_vmm(VMM_MAGIC_ID)
  if not equal(getn_vmm(4),"VMM"&VMM_VERSION) then
    fnActive = lastActive
    close(fn)
    return VMM_ERROR_INVALID
  end if
  go_get_nFree()
  fpFree = get4u_vmm()
  seek_vmm(fpFree-4)
  sizeFree = get4u_vmm()
  return fn
end function

global function create_vmm(string fname)
 integer fn
  fn = open(fname,"wb")
  if fn < 0 then return VMM_ERROR_OPEN end if
  close(fn)
  fn = open(fname,"ub")
  if fn < 0 then return VMM_ERROR_OPEN end if
  fnActive = fn
  putn_vmm("VMM")         -- magic id
  putn_vmm({VMM_VERSION}) -- version
  put4_vmm(0)             -- pointer to extra header
  put4_vmm(0)             -- number of free blocks
  put4_vmm(VMM_HEADER_SIZE+4) -- pointer to free blocks list
  put4_vmm(HEAP_STEP)           -- size of free blocks list
  putn_vmm(repeat(0,HEAP_STEP)) -- create the data space for the free blocks list
  return select_vmm(fn)
end function

---------------------------------------------------------------------------------
-- ALLOCATION ROUTINES --
procedure remove_entry(integer n)
 sequence tmp
  -- remove the next free block from the block list
  tmp = getn_vmm((nFree-n)*8) -- fetch the remainder of the Free Block List Entries to a buffer
  seek_vmm(fpFree+((n-1)*8)) -- go to the entry in the list
  putn_vmm(tmp) -- rewrite the buffer
  -- update the free block count
  go_put_nFree(nFree-1)
end procedure

procedure insert_entry(integer n, integer sizeX, integer fpX)
 sequence tmp
  seek_vmm(fpFree+((n-1)*8)) -- go to the entry in the list
  tmp = getn_vmm((nFree-n+1)*8) -- fetch the remainder of the Free Block List Entries to a buffer
  seek_vmm(fpFree+((n-1)*8)) -- go to the entry in the list
  put4_vmm(sizeX)   put4_vmm(fpX) -- insert new entry
  putn_vmm(tmp) -- rewrite the buffer
  -- update the free block count
  go_put_nFree(nFree+1)
end procedure

function recycle_vmm(integer len)
 integer size,fp,n

  if not nFree then return FALSE end if

  fp = FALSE
  n = 1
  seek_vmm(fpFree)
  for i = 1 to nFree do
    size = get4u_vmm()
    if size >= len then
      fp = get4u_vmm()
      n = i
      exit
    end if
    size = getc(fnActive)
    size = getc(fnActive)
    size = getc(fnActive)
    size = getc(fnActive)
  end for
  if not fp then return fp end if -- no block large enough was found

  -- see if we can segment it
  if size >= (len+12) then -- oversized block (at least enough space for an 8 byte data block)
    size -= (len+4)
    -- update the size entry in the free block list
    seek_vmm(fpFree+((n-1)*8))
    put4_vmm(size) -- update the size
    -- update the size in the free block
    seek_vmm(fp-4)    put4_vmm(size)
    -- update the size in the new block
    fp += size+4
    seek_vmm(fp-4)    put4_vmm(len)
  else -- not enough space to split, just activate it
    remove_entry(n)
    seek_vmm(fp) -- go to the data block
  end if

  return fp
end function

function append_vmm(integer size)
 integer fp
  seek_vmm(-1)
  fp = where(fnActive)+4
  put4_vmm(size)
  return fp
end function

global function alloc_vmm(sequence bytes)
 integer fp
 integer size

  size = align4(length(bytes))
  fp = recycle_vmm(size)
  if fp then
    putn_vmm(bytes)
  else
    fp = append_vmm(size)
    putn_vmm(bytes)
    if size != length(bytes) then
      putn_vmm(repeat(0,size-length(bytes)))
    end if
  end if
  return fp
end function

---------------------------------------------------------------------------------
-- DEALLOCATION ROUTINES --

global procedure dealloc_vmm(integer fpX)
 integer sizeX,sizeP,fpP,sizeN,fpN
 integer n
 sequence tmp

  seek_vmm(fpX-4)
  sizeX = get4u_vmm()

  n = 1
  seek_vmm(fpFree)
  while n <= nFree do
    sizeN = get4u_vmm() -- size
    fpN = get4u_vmm() -- pointer
    if fpN > fpX then
      exit
    end if
    sizeP = sizeN
    fpP = fpN
    n +=1
  end while

  if (n > 1) and (fpP+sizeP+4 = fpX) then -- combine with previous free block
    if (n < nFree) and (fpX+sizeX+4 = fpN) then -- combine with next free block
      sizeX += sizeN+4
      remove_entry(n)
    end if
    -- update the free block
    fpX = fpP
    sizeX += sizeP+4
    seek_vmm(fpX-4)
    put4_vmm(sizeX)
    -- update the entry in the free block list
    seek_vmm(fpFree+((n-2)*8))
    put4_vmm(sizeX)
  elsif (n <= nFree) and (fpX+sizeX+4 = fpN) then -- combine with next free block
    sizeX += sizeN+4
    seek_vmm(fpX-4)
    put4_vmm(sizeX)
    -- update the entry in the free block list
    seek_vmm(fpFree+((n-1)*8))
    put4_vmm(sizeX)    put4_vmm(fpX)
-- insert a new free block into the list
  elsif sizeFree = (nFree*8) then -- no room for more entries
    -- reallocate the free block list
    fpP = fpFree
    sizeFree +=HEAP_STEP
    fpFree = recycle_vmm(sizeFree)
    if fpFree then
      -- recycle_vmm() changes the Free Block count
      go_get_nFree()
    else
      fpFree = append_vmm(sizeFree)
      putn_vmm(repeat(0,sizeFree))
    end if
    dealloc_vmm(fpP)
    -- fetch the old Free Block List Entries to a buffer
    seek_vmm(fpP)
    tmp = getn_vmm(n*8)
    -- put the buffer to the new Free Block List
    seek_vmm(fpFree)
    putn_vmm(tmp)

    insert_entry(n,sizeX,fpX)
    put4_vmm(fpFree)
  else
    insert_entry(n,sizeX,fpX)
  end if
end procedure
